@@ -13,30 +13,49 @@
 # );
 #
 
+use IO::Handle;
+use Sys::Hostname;
 
 $dbg  = 0;
 $prog = "Signify";
-$vers = "1.01";
+$vers = "1.14";
+srand();
+
+use open ':locale';
+
+# If there is no $ENV{HOME}, set it from /etc/passwd.
+$ENV{HOME} = (getpwuid($<))[7] unless exists $ENV{HOME};
 
 
 %vars			= %ENV;
 $vars{PROGRAM}	= "$prog v$vers";
 $vars{SIGWIDTH}	= 79;
-$vars{WEBSITE}	= "http://www.verisim.com/";
-$errorfile		= "";
-$outputfifo		= "";
-$inputfile		= "$ENV{HOME}/.signify";
+$vars{WEBSITE}	= "http://www.debian.org/";
+$vars{DOLSIGN}	= "\$";
+$vars{BLANK}	= "";
+$perlvarnam		= "_perl_";
+$perlvarnum		= 1;
+$CurFile		= "";
+$CurPath		= "";
+@OpenFiles		= ();
+@OpenPaths		= ();
+$OutputFifo		= "";
+$InputFile		= "$ENV{HOME}/.signify";
+$FifoName		= undef;
+$FifoLock		= undef;
+$KillOther		= undef;
 
 
 sub Error {
-	die "$errorfile:$.:$_[0]\n";
+	die "$CurFile:$.:$_[0]\n";
 }
 
 
 
 sub SubstituteVars {
 	my($line) = @_;
-	$line =~ s/\$(\w+)/$vars{$1}/gs;
+	$line =~ s/\$\$/\$DOLSIGN;/gs;
+	$line =~ s/\$(\w+);?/$vars{$1}/gs;
 	return $line;
 }
 
@@ -59,10 +78,63 @@ sub DefaultSection {
 
 
 
+sub OpenFile {
+	my($file) = @_;
+
+	if ($CurPath || $CurFile) {
+		push @OpenPaths,$CurPath;
+		push @OpenFiles,$CurFile;
+	}
+
+	if (grep(/^\Q$file\E$/,@OpenFiles)) {
+		die "$prog: File '$file' is read recursively\n";
+	}
+
+	$CurFile = $file;
+	$CurPath = new IO::Handle;
+	open($CurPath,"<$file") || die "$prog: Could not read file '$file' -- $!\n";
+	print STDERR "Starting file '$CurFile'\n" if $dbg;
+}
+
+sub ReadLine {
+	my $line = <$CurPath>;
+
+	while (!defined $line || $line eq "") {
+		print STDERR "Finished file '$CurFile'\n" if $dbg;
+		close($CurPath);
+		$CurPath = pop @OpenPaths;
+		$CurFile = pop @OpenFiles;
+		last if !defined $CurPath || !defined $CurFile;
+		$line = <$CurPath>;
+	}
+
+	print STDERR "$CurFile: $line" if $dbg && $CurFile;
+	$line =~ s/^\#.*$//;
+	if ($line =~ m/\`([^\`]+?)\`/) {
+		my $pvar = "$perlvarnam$perlvarnum";
+		$line =~ s/\`([^\`]+?)\`/\$$pvar;/g;
+		$vars{$pvar} = eval $1;
+		$perlvarnum++;
+	}
+
+	if ($line =~ m/^\%\s*include\s+[\'\"]?(.*?)[\"\']?$/i) {
+		my $newfile;
+		$newfile = SubstituteVars($1);
+		$newfile =~s!^~/!$ENV{HOME}/!;
+		OpenFile($newfile);
+		$line = ReadLine();
+	}
+
+	return $line;
+}
+
+
+
 sub ReadCommand {
 	my($line) = @_;
 	if ($line =~ m/^\%\s*\$(\w+)=(.*)$/) {
 		$vars{$1} = SubstituteVars($2);
+		print STDERR "var '$1' = '$vars{$1}'\n" if $dbg;
 		return 1;
 	}
 	return;
@@ -71,7 +143,7 @@ sub ReadCommand {
 
 
 sub ReadSection {
-	my($path,$line) = @_;
+	my($line) = @_;
 	my %section = DefaultSection();
 	my $secttype;
 
@@ -91,15 +163,15 @@ sub ReadSection {
 		if (/^top$/i)			{ $section{valign}	= -1; next; }
 		if (/^vcenter$/i)		{ $section{valign}	=  0; next; }
 		if (/^weight=(.*)$/i)	{ $section{weight}	= $1; next; }
+		if (/^exec$/i)			{ $section{exec}	=  1; next; }
 
 		Error "Unknown option '$_'";
 	}
 
 	while (1) {
-		$line = <$path>;
+		$line = ReadLine();
 		Error "Unexpected end-of-file" unless $line;
 		chomp $line;
-		print STDERR "Reading (file) line: $line\n" if $dbg;
 		next if !$line || $line =~ m/^\#/;
 
 		if ($line =~ m/^\%/) {
@@ -111,12 +183,11 @@ sub ReadSection {
 #			Error "Cannot nest 'alternate' sections" if $line =~ m/^\%\s*\{/ && !$section{colflag};
 
 			if ($line =~ m/^\%\s*[\{\(]/) {
-				push @{$section{lines}},ReadSection($path,$line);
+				push @{$section{lines}},ReadSection($line);
 				next;
-			}
-			if ($line =~ m/^\%\s*\|/) {
+			} elsif ($line =~ m/^\%\s*\|/) {
 				$line =~ s/\|/$secttype/;
-				$section{next} = ReadSection($path,$line);
+				$section{next} = ReadSection($line);
 				return \%section;
 			}
 
@@ -124,6 +195,7 @@ sub ReadSection {
 
 			Error "Unknown command";
 		} else {
+			print STDERR "Adding (section) line '$line'\n" if $dbg;
 			push @{$section{lines}},SubstituteVars($line);
 		}
 	}
@@ -133,21 +205,18 @@ sub ReadSection {
 
 sub ReadFile {
 	my($file) = @_;
-	my $section = DefaultSection();
+	my $handle  = new IO::Handle;
+	my $oldfile = $CurFile;
+	my %section = DefaultSection();
 
-	$section{halign} = -1;
-	$section{valign} = -1;
+	OpenFile($file);
 
-	$errorfile=$file;
-	open(FILE,"<$file") || die "Error: Could not read file '$file' -- $!\n";
-
-	while (<FILE>) {
+	while ($_ = ReadLine()) {
 		chomp;
-		next if !$_ || /^\#/;
-		print STDERR "Reading (file) line: $_\n" if $dbg;
+		next if !$_;
 
 		if (/^\%\s*[\{\(]/) {
-			push @{$section{lines}},ReadSection(FILE,$_);
+			push @{$section{lines}},ReadSection($_);
 			print STDERR "finished reading section\n" if $dbg;
 		} elsif (/^\%/) {
 			Error "Unknown command" unless ReadCommand($_);
@@ -157,8 +226,6 @@ sub ReadFile {
 		}
 	}
 
-	close FILE;
-
 	return \%section;
 }
 
@@ -167,6 +234,7 @@ sub ReadFile {
 sub GenerateSig {
 	my($sect) = @_;
 	my(@lines,$line,$i);
+	my($status,$signal);
 
 	if ($$sect{colflag}) {
 
@@ -184,14 +252,33 @@ sub GenerateSig {
 			push @cols,\@clines;
 
 			foreach (@{$$csect{lines}}) {
-				if (ref) {
+			    if (ref) {
 					print STDERR "creference\n" if $dbg;
 					push @clines,GenerateSig($_);
-				} else {
+			    } elsif ($$csect{exec}) {
+					print STDERR "Executing shell command '$_'\n" if $dbg;
+					@output = `$_`;
+					$status = $? >> 8;
+					$signal = $? & 255;
+					if ($?) {
+						print STDERR ("Shell command returned $status exit status",
+									  $signal ? " (killed on signal $signal)" : "",
+									  "\n") if $dbg;
+					} else {
+						chomp @output;
+						if ($dbg) {
+							foreach (@output) {
+								print STDERR "Shell command output '$_'\n";
+							}
+						}
+						push @clines,@output;
+					}
+			    } else {
 					print STDERR "cline '$_'\n" if $dbg;
 					push @clines,$_;
-				}
-			}
+			    }
+ 			}
+
 			foreach (@clines) {
 				my $length = length;
 				$count++;
@@ -303,7 +390,7 @@ sub GenerateSig {
 				$sptr   = $$sptr{next};
 			}
 
-			my $index = rand() * $total;
+			my $index = rand($total);
 
 			while ($$sect{weight} < $index) {
 				$index -= $$sect{weight};
@@ -315,7 +402,25 @@ sub GenerateSig {
 			if (ref) {
 				print STDERR "reference\n" if $dbg;
 				push @lines,GenerateSig($_);
-			} else {
+		    } elsif ($$sect{exec}) {
+				print STDERR "Executing shell command '$_'\n" if $dbg;
+				@output = `$_`;
+				$status = $? >> 8;
+				$signal = $? & 255;
+				if ($?) {
+					print STDERR ("Shell command returned $status exit status",
+								  $signal ? " (killed on signal $signal)" : "",
+								  "\n") if $dbg;
+				} else {
+					chomp @output;
+					if ($dbg) {
+						foreach (@output) {
+							print STDERR "Shell command output '$_'\n";
+						}
+					}
+					push @lines,@output;
+				}
+		    } else {
 				print STDERR "line '$_'\n" if $dbg;
 				push @lines,$_;
 			}
@@ -327,7 +432,8 @@ sub GenerateSig {
 
 			for ($i=0; $i < @lines; $i++) {
 				my $length = length($lines[$i]);
-				$lines[$i] = " " x ($center - $length/2) . $lines[$i];
+				my $prefix = int($center - $length/2);
+				$lines[$i] = " " x $prefix . $lines[$i];
 			}
 		}
 		if ($$sect{halign} == -2) {
@@ -346,27 +452,51 @@ sub GenerateSig {
 
 sub MakeFifo {
 	my ($fifo) = @_;
+	$fifo =~ s!^~/!$ENV{HOME}/!;
 
-	die "Error: '$fifo' already exists and is not a FIFO\n" if (-e $fifo && ! -p $fifo);
+	$FifoName = $fifo;
+	die "$prog: '$fifo' already exists and is not a FIFO\n" if (-e $fifo && ! -p $fifo);
 	system("mkfifo -m 644 $fifo") if (! -p $fifo);
-	die "Error: Could not make FIFO '$fifo'\n" if (! -p $fifo);
+	die "$prog: Could not make FIFO '$fifo'\n" if (! -p $fifo);
 }
 
 
 
 sub LockFifo {
 	my($fifo) = @_;
-	my $lock = "$fifo.lock";
+	my $pid;
 
-	if (open(LOCK,"<$lock")) {
-		my $pid = <LOCK>;
-		chomp $pid;
-		kill 1,$pid;
+	$FifoLock = "$fifo.lock-$hostnam";
+
+
+	if (! -e $FifoLock) {
+		open(LOCK,">>$FifoLock") || die "$prog: Could not create file '$FifoLock' -- $!\n";
 		close(LOCK);
 	}
 
-	open(LOCK,">$lock") || die "Error: Could not write file '$lock' -- $!\n";
-	print LOCK $$;
+	open(LOCK,"+<$FifoLock") || die "$prog: Could not update file '$FifoLock' -- $!\n";
+	flock(LOCK,2) || die "$prog: could not lock file '$FifoLock' -- $!\n";
+	$pid = <LOCK>;
+	chomp $pid;
+	if (!$KillOther && -d "/proc/self") {
+		if (open(CMDLINE,"</proc/$pid/cmdline")) {
+			$cmd = <CMDLINE>;
+			if ($cmd =~ m/signify/i) {
+				die "$prog: other signify already running (use --kill to replace it)\n";
+			}
+		}
+	} else {
+		kill 14,$pid if ($pid > 1 && $pid != $$);
+		sleep(1);
+	}
+
+	if (! -e $FifoLock) {
+		open(LOCK,">>$FifoLock") || die "$prog: Could not create file '$FifoLock' -- $!\n";
+		close(LOCK);
+	}
+
+	seek LOCK,0,0;
+	print LOCK $$,"\n";
 	close(LOCK);
 }
 
@@ -378,41 +508,67 @@ sub LockFifo {
 
 foreach (@ARGV) {
 	if (/^--fifo=(.*)$/) {
-		$outputfifo = $1;
-		MakeFifo($outputfifo);
+		$OutputFifo = $1;
+		MakeFifo($OutputFifo);
 		next;
 	}
 	if (/^--input=(.*)$/) {
-		$inputfile = $1;
+		$InputFile = $1;
+		next;
+	}
+	if (/^--kill$/) {
+		$KillOther = 1;
+		next;
+	}
+	if (/^--debug$/) {
+		$dbg = 1;
 		next;
 	}
 
-	die "Error: Unknown parameter '$_'\n";
+	die "$prog: Unknown parameter '$_'\n";
 }
 
 
-$sectref = ReadFile($inputfile);
-$nowtime = 840000000;
+$sectref = ReadFile($InputFile);
+$nowtime = 1000000000;
+$hostnam = hostname();
 
 
-if ($outputfifo) {
-	LockFifo($outputfifo);
+sub ExitProg
+{
+#	unlink $FifoName if ($FifoName);	# may be in use across multiple machines
+	unlink $FifoLock if ($FifoLock);
+	exit(0);
+}
+
+$SIG{ALRM}	= sub { exit(0); };
+$SIG{HUP}	= sub { $sectref = ReadFile($InputFile); };
+$SIG{INT}	= \&ExitProg;
+$SIG{KILL}	= \&ExitProg;
+$SIG{TERM}	= \&ExitProg;
+$SIG{ABRT}	= \&ExitProg;
+$SIG{QUIT}	= \&ExitProg;
+
+
+if ($OutputFifo) {
+	local $SIG{PIPE} = sub {};
+	LockFifo($OutputFifo);
 	while (1) {
-		srand(time() ^ $$);
-		utime $nowtime,$nowtime,$outputfifo;
-		open(FIFO,">$outputfifo") || die "Error: Could not write to '$outputfifo' -- $!\n";
+		utime $nowtime,$nowtime,$OutputFifo;
+		open(FIFO,">$OutputFifo") || die "$prog: Could not write to '$OutputFifo' -- $!\n";
 		@lines = GenerateSig($sectref);
 		foreach (@lines) {
+			s/\s+$// unless m/^--/;
 			print FIFO $_,"\n";
 		}
 		close(FIFO);
-		utime $nowtime,$nowtime,$outputfifo;
+		utime $nowtime,$nowtime,$OutputFifo;
 		sleep(1);
 	}
 } else {
-	srand(time() ^ $$);
 	@lines = GenerateSig($sectref);
 	foreach (@lines) {
+		s/\s+$// unless m/^--/;
 		print $_,"\n";
 	}
 }
